{ ******************************************************************* }
{ Turbo Pascal Version 6.0 -> 7.0    Large memory based sets - object }
{ ------------------------------------------------------------------- }
{ Copyright 1996 Roger E. Donais              <rdonais@southeast.net> }
{ ------------------------------------------------------------------- }

UNIT BITSET;
{-R-,S-,O+} { enabled by ide }

{ ******************************************************************* }
INTERFACE

USES OBJECTS;

CONST MAX_BITSET = 65520 * 8;

{ ------------------------------------------------------------------- }
{ Operation of iterators and some indexing methods ignore bit 0       }
{ So it is best if you work your set from 1 to n rather than 0 to n-1 }
{ ------------------------------------------------------------------- }

TYPE
     tpMemSet = ^tMemSet;
     tMemSet  = OBJECT (TObject)
         Siz:  Word;
         Map: pByteArray;

         Constructor Init(aMap: Pointer; aByteCount: Word);
         Procedure   NullSet;
         Procedure   FillRange(From, By: Longint);
         Procedure   ClearRange(From, By: Longint);
         Function    IsEmpty                                : Boolean;
         Function    Count                                  : Longint;
         Function    Contains(No: Longint)                  : Boolean;
         Function    Entry(VAR Idx: Longint; No: Longint)   : Boolean;
                     { returns TRUE w/Idx = bit position if 1 <= No <= x.Count    }
                     { Example: x.Include(2);                                     }
                     {          x.Include(4);                                     }
                     {          x.Include(8);                                     }
                     {          x.Entry(Idx, 4) returns FALSE, Only 3 bits in set }
                     {          x.Entry(Idx, 3) returns TRUE w/Idx = 8            )
                     { ---------------------------------------------------------- }
         Function    IndexOf(VAR Idx: Longint; No: Longint) : Boolean;
                     { returns TRUE w/Idx = set occurance if No is in set         }
                     { Example: x.Include(2);                                     }
                     {          x.Include(4);                                     }
                     {          x.Include(8);                                     }
                     {          x.IndexOf(Idx, 3) returns FALSE, bit 3 is NOT set }
                     {          x.Indexof(Idx, 4) returns TRUE w/Idx = 2          )
                     { ---------------------------------------------------------- }
         Function    Next(VAR Idx: Longint; No: Longint)    : Boolean;
         Function    Prev(VAR Idx: Longint; No: Longint)    : Boolean;
         Function    Matches(VAR Other: tMemSet)            : Boolean;

         Function    Include(No: Longint): Boolean; { Returns TRUE if     }
         Function    Exclude(No: Longint): Boolean; {   bit was changed   }
         Procedure   AtInsert(No: Longint);         { Insert bit at No }
         Procedure   AtDelete(No: Longint);         { Delete bit at No }

         Procedure   IncludeSet(VAR Other: tMemSet); { set OR  other      }
         Procedure   ExcludeSet(VAR Other: tMemSet); { set AND NOT(other) }
         Procedure   Intersect(VAR Other: tMemSet);  { set AND other      }
     End;

     tpBitSet = ^tBitSet;
     tBitSet  = OBJECT (tMemSet)
         Sav: pByteArray;

         Constructor Init(MaxEntry: Longint);
         Constructor Clone(VAR Other: tMemSet);
         Constructor InitRec(VAR Rec; RecSize: Word);
         Destructor  Done;                                          VIRTUAL;

         Procedure   Swap;        { Do Save if Sav=NIL, Else Exchange Map & Sav }
         Procedure   Save;          { Make a copy of Map at Sav                 }
         Procedure   Restore;       { Dispose Map, Set map = Sav and Sav = NIL  }
         Procedure   IncludeSave;   { merge into save w/ include then restore   }
         Procedure   ExcludeSave;   { merge into save w/ exclude then restore   }
         Procedure   IntersectSave; { merge into save w/ intersect then restore }
         PRIVATE
           Procedure MergeIntoSave(Mode: Integer);
     End;

{ ******************************************************************* }
IMPLEMENTATION

{ =================================================================== }
{           L o c a l   S u p p o r t   P r o c d u r e s             }
{ =================================================================== }

PROCEDURE Position; Assembler;
{ ------------------------------------------------------------------- }
{ Given a tMemSet.SELF in DS:SI and a Bit number in DI:AX             }
{ Return                                                              }
{     DX:AX      Bit Number                                           }
{     DS:SI      Byte vector                                          }
{        DI      Siz                                                  }
{        BX      Byte index                                           }
{        CF      TRUE if AX within range of vector, Else CF FALSE     }
{     if TRUE also return ....                                        }
{        CL      Bit index                                            }
{        CH      Bit mask                                             }
{ ------------------------------------------------------------------- }
ASM
      { Get BX=Byte Index and CL=Bit index }
      mov  DX, DI
      mov  BX, AX
      shr  DX, 1; rcr BX, 1;
      shr  DX, 1; rcr BX, 1;
      shr  DX, 1; rcr BX, 1;
      mov  CX, 7
      and  CX, AX               { CL == bit index }
      mov  CH, $80              { CH == bit mask  }
      shr  CH, CL
      cmp  DX, 1

      mov  DX, [ (tBitSet PTR DS:SI).Siz ]
      lds  SI, [ (tBitSet PTR DS:SI).Map ]
      jnc  @@9                  { DX <> 0 means position too high }

      cmp  BX, DX
      jnc  @@9
      xchg DX, DI
      stc
@@9:
END;


{ ------------------------------------------------------------------- }
CONST WITH_EXCLUDE   = -1;   { NOT }  { Sav := Sav - Map }
      WITH_INCLUDE   =  0;   { OR  }  { Sav := Sav + Map }
      WITH_INTERSECT = +1;   { AND }  { Sav := Sav * Map }

PROCEDURE tBitSet.MergeIntoSave(Mode: Integer); Assembler;
{ ------------------------------------------------------------------- }
ASM
      push DS
      lds  SI, [ Self ]
      mov  CX, [ SI.Siz ]
      les  DI, [ SI.Sav ]
      lds  SI, [ SI.Map ]

      cld
      cmp  [ Mode ], WITH_INTERSECT
      jz   @@3
      jc   @@2

      { EXCLUDE }
@@1:  LODSB
      not  AL
      and  [ ES:DI ], AL
      inc  DI
      LOOP @@1
      jmp  @@9

@@2:  { INCLUDE }
      LODSB
      or   [ ES:DI ], AL
      inc  DI
      LOOP @@2
      jmp  @@9

@@3:  { INTERSECT }
      LODSB
      and  [ ES:DI ], AL
      inc  DI
      LOOP @@3

@@9:  pop  DS
END;


PROCEDURE BitCountTable; Assembler;
{ ------------------------------------------------------------------- }
ASM
     Call @@1
     {-  0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F ----}
      db 0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4  {0}
      db 1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5  {1}
      db 1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5  {2}
      db 2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6  {3}
      db 1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5  {4}
      db 2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6  {5}
      db 2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6  {6}
      db 3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7  {7}
      db 1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5  {8}
      db 2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6  {9}
      db 2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6  {A}
      db 3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7  {B}
      db 2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6  {C}
      db 3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7  {D}
      db 3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7  {E}
      db 4, 5, 5, 6, 5, 6, 6, 7, 5, 6, 6, 7, 6, 7, 7, 8  {F}
@@1:  push CS
      pop  ES
      pop  BX
END;


{ =================================================================== }
{                 t M e m S e t   M e t h o d s                       }
{ =================================================================== }

CONSTRUCTOR tMemSet.Init(aMap: Pointer; aByteCount: Word);
{ ------------------------------------------------------------------- }
BEGIN
    tObject.Init;
    Map := aMap;
    Siz := aByteCount;
END;

{ =================================================================== }
{                 t B i t S e t   M e t h o d s                       }
{ =================================================================== }

CONSTRUCTOR tBitSet.Init(MaxEntry: Longint);
{ ------------------------------------------------------------------- }
VAR p: Pointer;
    i: Word;
BEGIN
    If (MaxEntry > MAX_BITSET) Then RunError(201);
    i := (MaxEntry+7) div 8;
    GetMem(p, i);
    tMemSet.Init(p, i);
    Sav := NIL;
    NullSet;
END;


CONSTRUCTOR tBitSet.Clone(VAR Other: tMemSet);
{ ------------------------------------------------------------------- }
VAR p: Pointer;
BEGIN
    GetMem(p, Other.Siz);
    Move(Other.Map^, p^, Siz);
    tMemSet.Init(p, Other.Siz);
    Sav := NIL;
END;


CONSTRUCTOR tBitSet.InitRec(VAR Rec; RecSize: Word);
{ ------------------------------------------------------------------- }
BEGIN
    If (MAX_BITSET < RecSize * 8) Then RunError(201);
    tObject.Init;
    Siz := RecSize;
    Sav := NIL;
    GetMem(Map, RecSize);
    Move(Rec, Map^, RecSize);
END;


DESTRUCTOR tBitSet.Done;
{ ------------------------------------------------------------------- }
BEGIN
    FreeMem(Map, Siz);
    If Sav <> NIL Then Begin
       FreeMem(Sav, Siz);
       Sav := NIL;
    End;
    tMemSet.Done;
END;


PROCEDURE tMemSet.NullSet; Assembler;
{ ------------------------------------------------------------------- }
ASM
     les  DI, [ Self      ]
     mov  CX, [ ES:DI.Siz ]
     les  DI, [ ES:DI.Map ]
     cld
     xor  AL, AL
     REP  STOSB
END;


PROCEDURE SetRange{(From, By: Longint)} ; NEAR; Assembler;
{ ------------------------------------------------------------------- }
TYPE  Frame = RECORD
         BPreg : Word;
         Caller: Pointer;
         Self  : Pointer;
         By    : Longint;
         From  : Longint;
      End;
      Local = RECORD
         Frame   : Word;
         Caller  : Word;
         FillType: Word;
      End;

CONST FillType =  4;   { Local Caller, FillType }
ASM
        push AX        { fill/clear mask provided by caller }
        push DS
        jmp  @Start

@Fill:  cmp  [BP-FillType.Byte],0
        jz   @Clear

        or   [ DI ], AL
        mov  AL,-1
        inc  DI
        retn

@Clear: not  AL
        and  [ DI ], AL
        mov  AL,0
        inc  DI
        retn

@Start: lds  SI, [ Frame(BP).Self ]
        mov  AX, [ WORD PTR Frame(BP).From +0 ]
        mov  DI, [ WORD PTR Frame(BP).From +2 ]
        call Position
        jnc  @@9

        add  AX, [ WORD PTR Frame(BP).By +0 ]
        adc  DX, [ WORD PTR Frame(BP).By +2 ]
        sub  AX, 1
        sbb  DX, 0                 { DX:AX == Ending bit number    }
        {!!! need to force upper limit }

        mov  CH, 7
        and  CH, AL
        shr  DX, 1; rcr AX, 1
        shr  DX, 1; rcr AX, 1
        shr  DX, 1; rcr AX, 1
        or   DX, DX
        jz   @@1

        mov  AX, DI
        xor  CH, CH
        { At this point BX, CL contains lower bound }
        {           and AX, CH contains upper bound }
@@1:    cmp  AX, BX
        jc   @@9                   { Exit if Upper is below Lower!    }

        mov  DX, AX
        mov  AX, DI
        cmp  AX, DX
        jnc  @@2                   { Skip if upper byte w/in range    }

        dec  AX                    { Reset to upper limit of vector   }
        mov  DX, AX
        xor  CH, CH

@@2:    lea  DI, [ SI+BX ]
        push DS
        pop  ES
        mov  AL, -1
        shl  AL, CL                { Drop bits below range            }
        shr  AL, CL
        mov  CL, CH
        push CX
        cmp  DX, BX
        jz   @@4                   { Skip if Upper equals Lower       }

        call @Fill                 { Update byte at lower bound       }
        mov  BX, DX
        lea  CX, [ SI+BX ]         { CX == offset of upper byte       }
        sub  CX, DI                { Number of whole bytes in range   }
        jcxz @@4

        REP  STOSB
        mov  AL,-1

@@4:    pop  CX
        inc  CL
        sub  CL, 8
        neg  CL
        shr  AL, CL
        shl  AL, CL

        call @Fill

@@9:    pop  DS
        pop  AX
END;


PROCEDURE tMemSet.FillRange(From, By: Longint); Assembler;
{ ------------------------------------------------------------------- }
ASM
    mov   AX, -1
    call  SetRange
END;


PROCEDURE tMemSet.ClearRange(From, By: Longint); Assembler;
{ ------------------------------------------------------------------- }
ASM
    mov   AX, 0
    call  SetRange
END;


FUNCTION tMemSet.IsEmpty: Boolean; Assembler;
{ ------------------------------------------------------------------- }
ASM
     push DS
     lds  SI, [ Self   ]
     mov  CX, [ SI.Siz ]
     les  DI, [ SI.Map ]
     cld  { Set Forward Scan Direction }
     xor  AX, AX
     REPE SCASB         { Search for non-zero entry                      }
     cmc
     adc  AX, AX
     pop  DS
END;


FUNCTION tMemSet.Contains(No: Longint): Boolean; Assembler;
{ ------------------------------------------------------------------- }
ASM
     push DS
     lds  SI, [ Self ]
     mov  AX, [ No.Word.0 ]
     mov  DI, [ No.Word.2 ]
     call Position
     mov  AL, 0
     jnc  @@9

     and  CH, [ BX+SI ]
     jz   @@9

     inc  AL

@@9: pop  DS
END;


Function tMemSet.Include(No: Longint): Boolean; Assembler;
{ ------------------------------------------------------------------- }
ASM
     push DS
     lds  SI, [ Self ]
     mov  AX, [ No.Word.0 ]
     mov  DI, [ No.Word.2 ]
     call Position
     mov  AL, 0
     jnc  @@9

     test [ BX+SI ], CH    { Get state of bit ... }
     jnz  @@9

     inc  AL               { TRUE, Bit was changed ... }
     or   [ BX+SI ], CH

@@9: pop  DS
END;


Function tMemSet.Exclude(No: Longint): Boolean; Assembler;
{ ------------------------------------------------------------------- }
ASM
     push DS
     lds  SI, [ Self ]
     mov  AX, [ No.Word.0 ]
     mov  DI, [ No.Word.2 ]
     call Position
     mov  AL, 0
     jnc  @@9

     test [ BX+SI ], CH    { Get state of bit ... }
     jz   @@9

     inc  AL               { TRUE, Bit was changed ... }
     not  CH
     and  [ BX+SI ], CH

@@9: pop  DS
END;


PROCEDURE tMemSet.AtInsert(No: Longint); Assembler;
ASM
     push DS
     lds  SI, [ Self ]
     mov  AX, [ No.Word.0 ]
     mov  DI, [ No.Word.2 ]
     call Position
     jnc  @@9
     sub  DI, BX
     add  SI, BX

     xor  AH, AH
     mov  AL, [SI]
     shl  AX, CL            { shift lower bits into AH }
     shl  AH, 1             { insert specified bit    }
     inc  CL
     shr  AX, CL
     mov  [SI], AL
     mov  CX, DI
     jmp  @2

@1:  inc  SI
     rcr  BYTE PTR [SI], 1
@2:  loop @1

@@9: pop  DS
END;


PROCEDURE tMemSet.AtDelete(No: Longint); Assembler;
ASM
     push DS
     lds  SI, [ Self ]
     mov  AX, [ No.Word.0 ]
     mov  DI, [ No.Word.2 ]
     call Position
     jnc  @@9
     add  SI, DI
     sub  DI, BX
     xchg CX, DI
     xor  AH, AH

@1:  dec  SI
     rcl  BYTE PTR [SI], 1
     loop @1

     mov  AL, [SI]
     xchg CX, DI
     rcl  AH, 1         { capture hanging bit }
     shl  AX, CL        { shift low bits into AH }
     rcr  AH, 1         { discard specified bit  }
     shr  AX, CL        { shift bits back to AL  }
     mov  [SI],AL
@@9: pop  DS
END;


PROCEDURE tMemSet.IncludeSet(VAR Other: tMemSet); Assembler;
{ ------------------------------------------------------------------- }
ASM
     push DS
     lds  DI, [ Self   ]
     mov  CX, [ DI.Siz ]
     les  DI, [ DI.Map ]
     lds  SI, [ Other  ]
     mov  DX, [ SI.Siz ]
     lds  SI, [ SI.Map ]
     cmp  CX, DX
     jc   @@1

     mov  CX, DX

@@1: cld  { Set Forward Scan Direction }

@@2: lodsb                { Get other          }
     or   [ ES:DI ], AL   { Include w/ ourself }
     inc  DI
     LOOP @@2

@@9: pop  DS
END;


PROCEDURE tMemSet.ExcludeSet(VAR Other: tMemSet); Assembler;
{ ------------------------------------------------------------------- }
ASM
     push DS
     lds  DI, [ Self   ]
     mov  CX, [ DI.Siz ]
     les  DI, [ DI.Map ]
     lds  SI, [ Other  ]
     mov  DX, [ SI.Siz ]
     lds  SI, [ SI.Map ]
     cmp  CX, DX
     jc   @@1

     mov  CX, DX

@@1: cld  { Set Forward Scan Direction }

@@2: lodsb                { Get other           }
     not  AL
     and  [ ES:DI ], AL   { Mask out w/ ourself }
     inc  DI
     LOOP @@2

@@9: pop  DS
END;


PROCEDURE tMemSet.Intersect(VAR Other: tMemSet); Assembler;
{ ------------------------------------------------------------------- }
ASM
     push DS
     lds  DI, [ Self   ]
     mov  CX, [ DI.Siz ]
     les  DI, [ DI.Map ]
     lds  SI, [ Other  ]
     mov  DX, CX
     sub  DX, [ SI.Siz ]
     lds  SI, [ SI.Map ]
     jnc  @@1

     xor  DX, DX        { Other set is same size or larger }

@@1: cld   { Set Forward Scan Direction }

@@2: lodsb                { Get other           }
     and  [ ES:DI ], AL   { Mask out w/ ourself }
     inc  DI
     LOOP @@2

     mov  CX, DX
     jcxz @@9          { Other set is same size or larger }

     xor  AL, AL
     STOSB

@@9: pop  DS
END;


PROCEDURE tBitSet.Save;
{ ------------------------------------------------------------------- }
BEGIN Sav := Map;
      GetMem(Map, Siz);
      Move(Sav^, Map^, Siz);
END;


PROCEDURE tBitSet.Swap; Assembler;
{ ------------------------------------------------------------------- }
ASM
        les   DI, [ Self ]
        mov   AX, [ ES:DI.Sav.Word.2 ]
        or    AX, [ ES:DI.Sav.Word.0 ]
        jz    @2

        mov   AX, [ ES:DI.Sav.Word.2 ]
        xchg  AX, [ ES:DI.Map.Word.2 ]
        mov   [ ES:DI.Sav.Word.2 ], AX

        mov   AX, [ ES:DI.Sav.Word.0 ]
        xchg  AX, [ ES:DI.Map.Word.0 ]
        mov   [ ES:DI.Sav.Word.0 ], AX
        jmp   @Exit

        @2:
        push    ES
        push    DI
        call    tBitset.Save
@Exit:
END;


PROCEDURE tBitSet.Restore;
{ ------------------------------------------------------------------- }
BEGIN If Sav <> NIL Then Begin
         FreeMem(Map, Siz);
         Map := Sav;
         Sav := NIL;
      End;
END;


PROCEDURE tBitSet.IncludeSave;
{ ------------------------------------------------------------------- }
BEGIN MergeIntoSave(WITH_INCLUDE);
      Restore;
END;


PROCEDURE tBitSet.ExcludeSave;
{ ------------------------------------------------------------------- }
BEGIN MergeIntoSave(WITH_EXCLUDE);
      Restore;
END;

PROCEDURE tBitSet.IntersectSave;
{ ------------------------------------------------------------------- }
BEGIN MergeIntoSave(WITH_INTERSECT);
      Restore;
END;


FUNCTION tMemSet.Count: Longint; Assembler;
{ ------------------------------------------------------------------- }
ASM
     push DS
     lds  SI, [ Self   ]
     mov  CX, [ SI.Siz ]
     lds  SI, [ SI.Map ]

     xor  AX, AX
     xor  DI, DI
     xor  DX, DX
     call BitCountTable
     cld

@@1: LODSB
     SEGES XLAT
     add  DI, AX
     adc  DX, 0
     LOOP @@1

     mov  AX, DI
     pop  DS
END;


FUNCTION tMemSet.Entry(VAR Idx: Longint; No: Longint): Boolean; Assembler;
{ ------------------------------------------------------------------- }
ASM
     push DS
     lds  SI, [ Self   ]
     mov  CX, [ SI.Siz ]
     mov  DI, [ No.Word.0 ]
     mov  DX, [ No.Word.2 ]
     mov  AX, DX                { When No == ZERO, always return false }
     or   AX, DI
     jz   @@9

     lds  SI, [ SI.Map ]

     call BitCountTable
     xor  AX, AX
     push SI
     cld
     sub  DI, 1
     sbb  DX, 0

@@1: LODSB
     SEGES XLAT
     sub  DI, AX
     sbb  DX, 0
     jc   @@2
     LOOP @@1

     pop  DX

@@x: xor AX, AX
     jz  @@9

@@2: pop DX
     add  DI, AX          { restore bad count            }
     dec  SI              { step back to previous entry  }
     mov  CL, [ SI ]      { Recover bit map byte         }

     sub  SI, DX          { 8 * (Ofs(Here) - Ofs(Start)) }
     mov  AX, 8
     mul  SI

@@3: shl  CL, 1
     sbb  DI, 0
     jc   @@8

     add  AX, 1
     adc  DX, 0
     jmp  @@3

@@8: lds  SI, [ Idx ]
     mov  [ SI +0 ], AX
     mov  [ SI +2 ], DX
     mov  AX, 1

@@9: pop  DS
END;


FUNCTION tMemSet.IndexOf(VAR Idx: Longint; No: Longint): Boolean; Assembler;
{ ------------------------------------------------------------------- }
ASM
     push DS
     lds  SI, [ Self ]
     mov  AX, [ No.Word.0 ]
     mov  DI, [ No.Word.2 ]
     call Position
     mov  AX, 0
     jnc  @@9

     and  CH, [ BX+SI ]
     jz   @@9

     push BX
     lea  SI, [ SI+BX ]                 { Count from byte position }
     call BitCountTable
     xor  DH, DH
     mov  AL, [ SI ]
     xor  CL, 7
     shr  AL, CL                        { Discard higher bits           }
     xor  AH, AH                        { Normalize applicable bits     }
     xor  DX, DX                        { Clear result accumulator      }
     xor  DI, DI
     pop  CX
     inc  CX
     dec  SI
     std

@@1: SEGES XLAT
     add   DI, AX
     adc   DX, 0
     LODSB
     LOOP  @@1

@@8: lds  SI, [ Idx ]
     mov  [ SI +0 ], DI
     mov  [ SI +2 ], DX
     mov  AX, 1

@@9: pop  DS
END;


FUNCTION tMemSet.Next(VAR Idx: Longint; No: Longint): Boolean; Assembler;
{ ------------------------------------------------------------------- }
ASM
     push DS
     lds  SI, [ Self   ]
     mov  AX, [ No.Word.0 ]
     mov  DI, [ No.Word.2 ]
     call Position
     jnc  @@x             { Exit if byte index is out of range          }

     mov  CH, [ BX+SI ]   { Get the byte containing the caller's No.    }
     inc  CL
     shl  CH, CL
     jnz  @@4

     mov  CX, DI
     sub  CX, BX          { Remaining bytes within the vector...        }
     dec  CX
     lea  DI, [ BX+SI+1 ] { Address the byte beyond the caller's entry  }

     push DS
     pop  ES
     cld
     xor  AL,AL
     REPE SCASB        { Locate the next NON-Zero byte                }
     jnz  @@2

@@x: xor  AX, AX
     jz   @@9          { short jump as Z has been guaranteed...      }

@@2: dec  DI           { Recover non-zero byte's address              }
     mov  CH, [ DI ]   { Recover the non-zero byte from the vector    }

     sub  DI, SI       { Compute the staring bit number               }
     mov  AX, 8
     mul  DI

@@3: shl  CH, 1
     jc   @@8

@@4: add  AX, 1
     adc  DX, 0
     jmp  @@3

@@8: lds SI, [ Idx ]
     mov [ SI +0 ], AX
     mov [ SI +2 ], DX
     mov AX, 1

@@9: pop  DS
END;


FUNCTION tMemSet.Prev(VAR Idx: Longint; No: Longint): Boolean; Assembler;
{ ------------------------------------------------------------------- }
ASM
     push DS
     lds  SI, [ Self   ]
     mov  AX, [ No.Word.0 ]
     mov  DI, [ No.Word.2 ]
     call Position
     jc   @@1

     mov  CX, DI
     mov  DI, SI          { Begin scan with the last byte of the vector }
     add  DI, CX
     jmp  @@2

@@x: xor  AX, AX
     jz   @@9             { short jump as Z has been guaranteed...      }

@@1: lea  DI, [ BX+SI ]   { Get the byte containing the caller's entry  }
     push BX
     xor  BL, BL
     mov  BL, [ DI ]      { Get the byte containing the caller's No.    }
     shl  BX, CL          { and move lower bits into DL                 }
     or   BH, BH
     pop  CX
     jnz  @@3

@@2: std                  { Scan in reverse }
     dec  DI
     push DS
     pop  ES
     xor  AL, AL
     REPE SCASB           { Locate non-zero byte  }
     jz   @@x             { Exit if nothing found }

     inc  DI              { Recover address of the non-zero byte     }
     mov  BH, [ DI ]      { Recover non-zero byte                    }
     sub  DI, SI          { Compute bit number one beyond this entry }
     mov  AX, 8
     inc  DI
     mul  DI

@@3: sub  AX, 1
     sbb  DX, 0
     shr  BH, 1
     jnc  @@3

@@8: lds  SI, [ Idx ]
     mov  [ SI +0 ], AX
     mov  [ SI +2 ], DX
     mov  AX, 1

@@9: pop  DS
END;


FUNCTION tMemSet.Matches(VAR Other: tMemSet): Boolean; Assembler;
{ ------------------------------------------------------------------- }
ASM
     push DS
     xor  BX, BX
     lds  SI, [ Other  ]
     mov  DX, [ SI.Siz ]
     les  DI, [ SI.Map ]
     lds  SI, [ Self   ]
     mov  CX, [ SI.Siz ]
     lds  SI, [ SI.Map ]

     cld
     cmp  CX, DX
     jz   @@3

     push DS
     push SI
     push ES
     push DI
     jc   @@2

     { Other set is larger }
     push  ES
     pop   DS
     mov   DI, SI
     xchg  CX, DX

@@2: { set in DS:SI, CX is larger }

     sub   CX, DX
     mov   DI, SI
     add   DI, DX
     xor   AL, AL
     REPE  SCASB

     mov   CX, DX
     pop   DI
     pop   ES
     pop   SI
     pop   DS
     jnz   @@9

@@3: REP  CMPSB
     jnz  @@9

     inc  BX

@@9: mov  AX, BX
     pop  DS
END;


END.
{ ------------------------------------------------------------------- }
{ Following was used to build the BitCountTable ...                   }
{ ------------------------------------------------------------------- }
VAR F: Text;
    i, j, k: Integer;

BEGIN
    Assign(F, 'JK'); Rewrite(F);
    For i := 0 To 255 Do Begin
        Case (i and $F) Of
           0: Write(F,'   db');
          Else Write(F,',');
        End;
        j := i;
        k := 0;
        While j <> 0 Do Begin
            Inc(k, Ord(Odd(j)));
            j := j shr 1;
        End;
        Write(F, k:2);
        If (i and $F) = 15 Then Writeln(F, '  {', i shr 4, '}');
    End;
    Close(F);
END.

